{$A+,B-,D+,E-,F-,G+,I-,L+,N-,O-,R-,S-,V+,X+}
{$M 16384,0,655360}

{ --------------------------------------------------------------------------- }
{ TIMER.PAS  Low-level routines to handle hardware timer interrupt and to     }
{            program hardware clock.                                          }
{            coolTimer Version 1.21a/P                                        }
{                                                                             }
{ Copyright(c) 1993-96 by B-coolWare.  Written by Bobby Z.                    }
{ --------------------------------------------------------------------------- }
{ History:

            Mar 13, 1993  - v1.00
                          initially written.
            Nov 9,  1993  - v1.05
                          added system clock independance, interface slightly
                          changed, new routines added to simplify operation.
            Nov 30, 1993  - v1.06
                          added ExitProc overriding. You do not have to call
                          ResetTimer before exiting anymore - ExitProc will do
                          it by itself.
            Jun 22, 1994  - v1.07
                          changed timer adjust logic - there was a way I
                          missed...
            Jan 27, 1995  - v1.10
                          added two more threads. Still quite precise though.
            Jun 13, 1995  - v1.11
                          C and Pascal versions routine names unified.
            Feb 19, 1996  - v1.20
                          Added more channels, simplified channel addition.
                          More compatibility with multitaskers added (using
                          only documented DOS functions) though I don't think
                          they will allow for precision.
            May 30, 1996  - v1.21
                          Fixed one small bug that rendered the system
                          inactive if a handler was assigned a rate less than
                          or equal to 18 TpS AFTER a handler with faster rate
                          was activated.
                          Thanx goes to eXORcist who discovered the bug.
            Jun 27, 1996  - v1.21a
                          removed one far keyword in C version...
*** end history }

unit Timer;

interface

const
        ctOk                  = 0;
        ctInvalidChannelIndex = -1;
        ctInternalError       = -666;

        ChannelCount          = 5;      { change this to increase/decrease }
                                        { number of available channels }
                                        { I don't recommend to set this value }
                                        { more than 50 - not sure it will work}
                                        { with so many threads. }

procedure setTimerDelay( TDelay : Word );
{ set timer latch to specified value. Used in conjunction with ComputeDelay  }

procedure setTimerSpeed( TpS : Word );
{ set timer speed to TpS ticks per second.                                   }
{ this won't let you set it to value below 18.2 ticks per second             }

procedure normTimerSpeed;
{ set timer speed to 18.2 ticks per second (default system value)            }

procedure resetTimer;
{ this will reset timer speed and internal handler. Call before termination  }
{ of your program.                                                           }

procedure disableHandler( Channel : Word );
{ this will reset custom handler. You should call SetTimerHandler to reenable}
{ custom handler.                                                            }

function setTimerHandler( Channel, Rate : Word; Handler : Pointer ) : Integer;
{ set Channel'th handler address to Handler and its rate to Rate.            }
{ Handler should be FAR routine without parameters. In should preserve SS,   }
{ SP, BP and DS registers. All other registers can be freely modified.       }

procedure changeRate( Channel, Rate : Word );
{ changes polling rate for given channel. If channel is not active, this     }
{ routine does nothing.                                                      }

procedure grabTimer;
{ should be called prior to any calls to other routines in this module except}
{ ComputeDelay and SetTimerDelay.                                            }

function ComputeDelay ( TpS: Word ) : Word;
{ Argument is value of how much times per second timer int should be gen'd.  }
{ Returns value to be passed to SetTimerDelay.                               }

procedure playSound( Hz : Word );

procedure noSound;

implementation

type
      TChannel = record
                  cHandler : pointer;   { routine to call }
                  cRate    : Word;      { polling rate }
                  cTicks   : Word;      { current ticks passed }
                end;

const
      Save08        : pointer = nil;
      OldExitProc   : pointer = nil;
      TicksPerClk   : Word = 0;         { to maintain system clock }
      CurrentTicks  : Word = 0;         { --"-- }
      Channels      : array[0..ChannelCount-1] of TChannel =
                      ((cHandler:nil; cRate:0; cTicks:0),
                      (cHandler:nil; cRate:0; cTicks:0),
                      (cHandler:nil; cRate:0; cTicks:0),
                      (cHandler:nil; cRate:0; cTicks:0),
                                  { ... }
                      (cHandler:nil; cRate:0; cTicks:0));

procedure setTimerSpeed; assembler;
asm
        mov     CurrentTicks,0
        push    TpS
        call    ComputeDelay
        mov     TicksPerClk,ax
        push    ax
        call    SetTimerDelay
end;

procedure setTimerDelay; assembler;
asm
        cli
        mov     al,36h
        out     43h,al
        mov     ax,TDelay
        out     40h,al
        xchg    ah,al
        out     40h,al
        sti
end;

procedure internalHandler; interrupt; forward;

procedure grabTimer; assembler;
asm
        sub     ax,ax
        mov     CurrentTicks,ax
        mov     TicksPerClk,ax
        mov     ax,3508h
        int     21h
        mov     word ptr Save08,bx
        mov     word ptr Save08[2],es
        push    ds
        pop     es
        mov     di,offset Channels              { clear all records }
        mov     cx,(type TChannel)*ChannelCount
        sub     ax,ax
        rep     stosw
        push    ds
        push    seg internalHandler
        pop     ds
        mov     dx,offset internalHandler
        mov     ax,2508h
        int     21h
        pop     ds
end;

procedure resetTimer; assembler;
asm
        call    normTimerSpeed          { reset speed back to default }
        mov     ax,ChannelCount-1
@@1:
        push    ax
        push    ax
        call    disableHandler          { disable all active handlers one by }
        pop     ax                      { one...                             }
        dec     ax
        jns     @@1
        mov     si,offset Save08        { ...and restore old timer interrupt }
        mov     ax,word ptr [si]        { handler.                           }
        or      ax,word ptr [si+2]
        jz      @@Q
        push    ds
        mov     dx,word ptr Save08
        mov     ax,word ptr Save08[2]
        mov     ds,ax
        mov     ax,2508h
        int     21h
        pop     ds
@@Q:
end;

function setTimerHandler; assembler;
asm
        mov     si,Channel                      { check if channel no. is ok }
        cmp     si,ChannelCount-1
        mov     ax,ctInvalidChannelIndex
        ja      @@Q
        mov     ax,si
        mov     cx,type TChannel                { get si to point to given }
        mul     cl                              { channel structure        }
        mov     si,ax
        mov     ax,Rate
        push    ax
        call    ComputeDelay                    { compute ticks needed for }
        mov     TChannel(Channels[si]).cRate,ax { given rate and save them }
        mov     TChannel(Channels[si]).cTicks,ax
        mov     di,ax
        cli
        mov     ax,word ptr Handler
        mov     word ptr TChannel(Channels[si]).cHandler,ax
        mov     ax,word ptr Handler[2]
        mov     word ptr TChannel(Channels[si]).cHandler[2],ax
        sti
        mov     ax,TicksPerClk  { timer rate is not set yet? }
        test    ax,ax
        jnz     @@1             { already set - check if it should be changed }
@@2:
        push    Rate
        call    setTimerSpeed
        jmp     @@Ok
@@1:
        test    di,di   { is this standard rate? }
        jz      @@Ok    { yes - no need to reprogram the timer }
        cmp     di,ax   { do we need to reprogram timer for increased rate? }
        jb      @@2     { yes - do it }
@@Ok:
        mov     ax,ctOK
@@Q:
end;


procedure internalHandler; assembler;
 asm
        mov     ax,seg @data
        mov     ds,ax
        sub     si,si
@@1:
        {first handle all channels with non-standard rates }
        mov     ax,TChannel(Channels[si]).cRate
        test    ax,ax
        jz      @@2
        mov     bx,TicksPerClk
        sub     TChannel(Channels[si]).cTicks,bx
        ja      @@2
        add     TChannel(Channels[si]).cTicks,ax
        mov     ax,word ptr TChannel(Channels[si]).cHandler
        or      ax,word ptr TChannel(Channels[si]).cHandler[2]
        jz      @@2
        push    si
        call    dword ptr TChannel(Channels[si]).cHandler
        pop     si
 @@2:
        add     si,type TChannel
        cmp     si,(type TChannel)*ChannelCount
        jb      @@1
        jmp     @@Q
        db      13,10,'coolTimer v1.21/P  Copyright(c) 1993-96 by B-coolWare',13,10
 @@Q:
        mov     ax,TicksPerClk
        test    ax,ax
        jz      @@21
        add     CurrentTicks,ax
        jnc     @@11
 @@21:
        sub     si,si
 @@22:
        { handle all channels with system timer speed }
        mov     ax,TChannel(Channels[si]).cRate { not with normal speed? }
        test    ax,ax
        jnz     @@31
        mov     ax,word ptr TChannel(Channels[si]).cHandler
        or      ax,word ptr TChannel(Channels[si]).cHandler[2]
        jz      @@31
        push    si
        call    dword ptr TChannel(Channels[si]).cHandler
        pop     si
@@31:
        add     si,type TChannel
        cmp     si,(type TChannel)*ChannelCount
        jb      @@22
        { call system timer handler }
        pushf
        call    dword ptr [Save08]
        jmp     @@IRQHandled
 @@11:
        mov     al,20h
        out     20h,al
 @@IRQHandled:
end;

procedure changeRate;
 var Temp : Pointer;
 begin
  if Channels[Channel].cHandler <> nil then
   begin
    Temp := Channels[Channel].cHandler;
    disableHandler(Channel);
    setTimerHandler(Channel, Rate, Temp);
   end;
 end;

function ComputeDelay; assembler;
asm
        mov     bx,TpS
        cmp     bx,18
        mov     ax,0
        jbe     @@Q
        mov     dx,12h
        mov     ax,34DDh
        div     bx
@@Q:
end;

procedure playSound; assembler;
asm
        mov     bx,Hz
        mov     ax,34DDh
        mov     dx,0012h
        cmp     dx,bx
        jnc     @@2
        div     bx
        push    ax
        in      al,61h
        test    al,3
        jnz     @@1
        or      al,3
        out     61h,al
        mov     al,0B6h
        out     43h,al
@@1:
        pop     ax
        out     42h,al
        xchg    al,ah
        out     42h,al
@@2:
end;

procedure noSound; assembler;
asm
        in      al,61h
        and     al,0FCh
        out     61h,al
end;

procedure disableHandler; assembler;
asm
        mov     ax,Channel
        mov     cx,type TChannel
        mul     cl
        mov     si,ax
        sub     ax,ax
        mov     TChannel(Channels[si]).cTicks,ax
        mov     TChannel(Channels[si]).cRate,ax
        cli
        mov     word ptr TChannel(Channels[si]).cHandler,ax
        mov     word ptr TChannel(Channels[si]).cHandler[2],ax
        sti
end;

procedure normTimerSpeed; assembler;
asm
        sub     ax,ax
        push    ax
        push    ax
        call    setTimerDelay
        pop     ax
        mov     TicksPerClk,ax
        mov     CurrentTicks,ax
        sub     si,si
        mov     cx,ChannelCount
@@1:
        mov     TChannel(Channels[si]).cTicks,ax
        mov     TChannel(Channels[si]).cRate,ax
        add     si,type TChannel
        loop    @@1
end;

procedure coolTimerExit; far; assembler;
asm
        call    resetTimer
        les     bx,[OldExitProc]
        mov     word ptr ExitProc,bx
        mov     word ptr ExitProc[2],es
end;

begin
 OldExitProc := ExitProc;
 ExitProc := @coolTimerExit;
end.
